 ############################################################################
 #                                                                          #
 #                              ARCH.PY                                     #
 #                                                                          #
 #           Copyright (C) 2008 - 2014 Ada Core Technologies, Inc.          #
 #                                                                          #
 # This program is free software: you can redistribute it and/or modify     #
 # it under the terms of the GNU General Public License as published by     #
 # the Free Software Foundation, either version 3 of the License, or        #
 # (at your option) any later version.                                      #
 #                                                                          #
 # This program is distributed in the hope that it will be useful,          #
 # but WITHOUT ANY WARRANTY; without even the implied warranty of           #
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
 # GNU General Public License for more details.                             #
 #                                                                          #
 # You should have received a copy of the GNU General Public License        #
 # along with this program.  If not, see <http://www.gnu.org/licenses/>     #
 #                                                                          #
 ############################################################################

"""System information handling

This package contains a single class called Arch that allows the user to
instantiate configuration objects containing information about the system
(native or cross).
"""

# Export only the Arch class
__all__ = ['Arch']

import platform
import re
import os.path
from gnatpython import config
from gnatpython.ex import Run

UNKNOWN = 'unknown'

# The function that retrieve the linux release file is put apart from the
# class so that we can easily override it during testing.


def get_linux_release_file():
    """Get release file name and content

    RETURNS
      a tuple (name of the distrib info file, content of the file)
      or ('', '') is not found
    """
    for release_file in ('redhat-release', 'SuSE-release', 'lsb-release',
                         'debian_version'):
        if os.path.isfile('/etc/' + release_file):
            with open('/etc/' + release_file, 'r') as fd:
                content = fd.read()
            return (release_file, content)
    return ('', '')


class CPU(object):
    """CPU attributes

    ATTRIBUTES
      name: string containing the cpu name
      bits: int representing the number of bits for the cpu or 'unknown'
      endian: 'big', 'little' or 'unknown'
    """

    def __init__(self):
        self.name = UNKNOWN
        self.bits = UNKNOWN
        self.endian = UNKNOWN
        self.cores = UNKNOWN


class OS(object):
    """OS attributes

    ATTRIBUTES
      name: os name
      version: string containing the os version
      exeext: default executable extension
      dllext: default shared library extension
      is_bareboard: True if the system is bareboard, False otherwise
    """

    def __init__(self):
        self.name = UNKNOWN
        self.version = None
        self.exeext = ''
        self.dllext = ''
        self.is_bareboard = False
        self.kernel_version = None


class Arch(object):
    """Class that allow user to retrieve os/cpu specific informations

    ATTRIBUTES
      cpu:      CPU information (see _Arch__CPU)
      os:       Operating system information (see _Arch__OS)
      is_hie:   True if the system is a high integrity system
      platform: AdaCore platform product name. Ex: x86-linux
      triplet:  GCC target
      machine:  machine name
      domain:   domain name
      is_host:  True if this is not a cross context
      is_virtual: Set to True if the current system is a virtual one.
                  Currently set only for Solaris containers, Linux VMware
                  and Windows on VMware.
    """

    def __init__(self, platform_name=None, version=None, is_host=False,
                 machine=None):
        """Arch constructor

        PARAMETERS
          platform: if None or empty then automatically detect current platform
            (native) Otherwise should be a valid platform string.
          version:  if None, assume default OS version or find it automatically
            (native case only). Otherwise should be a valid version string.
          is_host:  if True the system is not a cross one. Default is False
            except if a platform_name is specified or if the platform_name is
            equal to the automatically detected one.

        RETURNS
          an Arch instance
        """

        # Create necesarry namespaces using "dummy" classes __CPU and __OS
        self.cpu = CPU()
        self.os = OS()

        # Initialize attributes values
        self.platform = platform_name or None  # force autodetection if empty
        self.os.version = version
        self.machine = machine
        self.is_hie = False
        self.is_virtual = False
        self.domain = UNKNOWN

        if self.platform is None:
            self.is_host = True
        else:
            self.is_host = is_host

        if self.platform is None:
            # In this case we try to guess the host platform
            self.platform = self.__guess_platform()
        else:
            if self.platform == self.__guess_platform():
                # This is a native platform
                self.is_host = True

        if self.is_host:
            # This is host so we can find the machine name using uname fields
            tmp = platform.uname()[1].lower().split('.', 1)
            self.machine = tmp[0]
            if len(tmp) > 1:
                self.domain = tmp[1]
            else:
                # Not found in uname(), try with the default NIS domain
                try:
                    import nis
                    try:
                        self.domain = nis.get_default_domain()
                    except nis.error:
                        pass
                except ImportError:
                    # Keep domain set to UNKNOWN.
                    pass

            # On solaris host detect if we are in a container context or not
            tmp = platform.uname()
            if tmp[0] == 'SunOS' and tmp[3] == 'Generic_Virtual':
                self.is_virtual = True
        else:
            # This is a target name. Sometimes it's suffixed by the host os
            # name. If the name is not a key in config.platform_info try to
            # to find a valid name by suppressing -linux, -solaris or -windows
            if self.platform not in config.platform_info:
                for suffix in ('-linux', '-solaris', '-windows'):
                    if self.platform.endswith(suffix):
                        self.platform = self.platform[:-len(suffix)]
                        break

        # Fill other attributes
        self.__fill_info()

        # Try to guess if we are on virtual machine. Don't do that check
        # earlier as we Run will not work otherwise.
        if self.is_host:
            if self.os.name == 'linux' and os.path.isfile('/sbin/ifconfig'):
                # On linux the start of the mac address can tell us
                # if this is a virtual machine running on VMware
                eth0 = Run(['/sbin/ifconfig', 'eth0']).out
                m = re.search(r'HWaddr (..:..:..):..:..', eth0, re.S)
                if m is not None:
                    mac_address = m.group(1)
                    if mac_address in ('00:0C:29', '00:50:56'):
                        self.is_virtual = True
            elif self.os.name == 'windows' and \
                    os.path.isfile('/WINDOWS/system32/ipconfig.exe'):
                # On windows if one of the ethernet interface name contains
                # VMware, then we are on a virtual machine.
                out = Run(['/WINDOWS/system32/ipconfig.exe', '/all']).out
                if 'VMware' in out:
                    self.is_virtual = True

        # Find triplet
        self.triplet = config.build_targets[self.platform]['name'] % \
            self.__get_dict()

    def __get_dict(self):
        """Export os and cpu variables as os_{var} and cpu_{var}

        Returns a dictionary containing os and cpu exported vars
        and self.__dict__ content
        """
        str_dict = self.__dict__.copy()
        for (key, var) in self.os.__dict__.items():
            str_dict["os_" + key] = var
        for (key, var) in self.cpu.__dict__.items():
            str_dict["cpu_" + key] = var
        return str_dict

    def __str__(self):
        """Return a representation string of the object"""
        result = "platform: %(platform)s\n" \
            "machine:  %(machine)s\n" \
            "is_hie:   %(is_hie)s\n" \
            "is_host:  %(is_host)s\n" \
            "is_virtual: %(is_virtual)s\n" \
            "triplet:  %(triplet)s\n" \
            "domain:   %(domain)s\n" \
            "OS\n" \
            "   name:          %(os_name)s\n" \
            "   version:       %(os_version)s\n" \
            "   exeext:        %(os_exeext)s\n" \
            "   dllext:        %(os_dllext)s\n" \
            "   is_bareboard:  %(os_is_bareboard)s\n" \
            "CPU\n" \
            "   name:   %(cpu_name)s\n" \
            "   bits:   %(cpu_bits)s\n" \
            "   endian: %(cpu_endian)s\n" \
            "   cores:  %(cpu_cores)s" % self.__get_dict()
        return result

    def __fill_info(self):
        """Internal function that fill info related to the cpu, os, ..."""
        self.os.name = config.platform_info[self.platform]['os']
        self.cpu.name = config.platform_info[self.platform]['cpu']
        self.is_hie = config.platform_info[self.platform]['is_hie']

        self.cpu.bits = config.cpu_info[self.cpu.name]['bits']
        self.cpu.endian = config.cpu_info[self.cpu.name]['endian']

        # Some platforms do not use the CPU default endianness. Check if there
        # is some override
        if 'endian' in config.platform_info[self.platform]:
            self.cpu.endian = config.platform_info[self.platform]['endian']

        self.os.is_bareboard = config.os_info[self.os.name]['is_bareboard']
        self.os.exeext = config.os_info[self.os.name]['exeext']
        self.os.dllext = config.os_info[self.os.name]['dllext']

        # If version is not given by the user guess it or set it to the
        # default (cross case)
        if self.is_host and self.os.version is None:
            self.__guess_os_version()
        if self.os.version is None:
            self.os.version = config.os_info[self.os.name]['version']
        if self.is_host:
            self.cpu.cores = self._compute_nb_cores()

    def __guess_platform(self):
        """Guess platform name

        Internal function that guess base on uname system call the
        current platform

        RETURNS
          a string object containing the platform name
        """

        def re_contains(left, right):
            """Returns right in left (regexp aware)"""
            if re.match(left + '$', right) or re.match(
                    '^' + left, right):
                return True
            else:
                return False

        def re_endswith(left, right):
            """Returns right.endswith(left) (regexp aware)"""
            return re.match(left + '$', right)

        def guess(os_name, p_uname):
            """Guess based on os_name"""
            for p_name in config.host_guess:
                p_config = config.host_guess[p_name]
                if p_config['os'] is not None:
                    if re_contains(p_config['os'], os_name):
                        if p_config['cpu'] is None or \
                                re_endswith(p_config['cpu'], p_uname[4]) or \
                                re_endswith(p_config['cpu'], p_uname[5]):
                            # The p_name config matched

                            if p_name in config.host_aliases:
                                return config.host_aliases[p_name]
                            else:
                                return p_name
            # wrong guess
            return None

        # Else we need to guess
        uname = platform.uname()

        p_name = guess(uname[0], uname)
        if p_name is not None:
            return p_name

        p_name = guess(uname[2], uname)
        if p_name is not None:
            return p_name

        # Not found !
        return UNKNOWN

    def __guess_os_version(self):
        """Internal function used to guess the host OS version/dist

        Set the self.os.version attribute and on some platform the
        self.os.kernel_version
        """
        uname = platform.uname()
        if self.os.name == 'darwin':
            self.os.version = uname[2]
        elif self.os.name == 'freebsd':
            self.os.version = re.sub('-.*', '', uname[2])
        elif self.os.name == 'linux':
            self.os.kernel_version = uname[2]
            linux_release = get_linux_release_file()

            if linux_release[0] == 'redhat-release':
                # RedHat distributions
                # Return the version as rhES4, rhES5, rhES6
                # ignoring the 'minor' version

                content = linux_release[1].strip()
                for sub in (('\(.*', ''),
                            (' ', ''),
                            ('Linux', ''),
                            ('release', ''),
                            ('Enterprise', ''),
                            ('AdvancedServer', 'AS'),
                            ('Server', 'ES'),
                            ('RedHat', 'rh'),
                            ('([0-9]+)\.[0-9.]+', r'\1')):
                    content = re.sub(sub[0], sub[1], content)
                self.os.version = content
            elif linux_release[0] == 'SuSE-release':
                # Suse distributions
                for line in linux_release[1].splitlines():
                    version = re.search('VERSION = ([0-9\.]+)', line)
                    if version is not None:
                        self.os.version = 'suse' + version.group(1)
                        break
                if self.os.version is None:
                    self.os.version = 'suse'
            elif linux_release[0] == 'lsb-release':
                # /etc/lsb-release is present on the previous distrib
                # but is not useful. On ubuntu it contains the
                # distrib number
                distrib_name = ''
                distrib_version = ''
                for line in linux_release[1].splitlines():
                    distrib_id = re.search('DISTRIB_ID=(.+)', line.rstrip())
                    if distrib_id is not None:
                        distrib_name = distrib_id.group(1).lower()
                    else:
                        distrib_release = re.search('DISTRIB_RELEASE=(.*)',
                                                    line.rstrip())
                        if distrib_release is not None:
                            distrib_version = distrib_release.group(1)

                if distrib_name:
                    self.os.version = distrib_name + distrib_version
            elif linux_release[0] == 'debian_version':
                self.os.version = 'debian_' + \
                    linux_release[1].strip().replace('/', '_')
        elif self.os.name == 'aix':
            self.os.version = uname[3] + '.' + uname[2]
        elif self.os.name == 'hp-ux':
            version = uname[2]
            if version[0:2] == 'B.':
                version = version[2:]
            self.os.version = version
        elif self.os.name == 'irix':
            self.os.version = uname[2]
        elif self.os.name == 'solaris':
            self.os.version = '2' + uname[2][1:]
        elif self.os.name == 'windows':
            self.os.version = uname[2].replace('Server', '')
            self.os.kernel_version = uname[3]
            if self.os.version == 'Vista' and '64' in uname[4]:
                self.os.version = 'Vista64'
            elif self.os.version == 'post2008':
                # Current version of Python is not able to find the
                # windows version for Windows 8 and Windows Server 2012
                # So find the version using the kernel version.
                is_server = 'Server' in uname[2]
                if self.os.kernel_version.startswith('6.2'):
                    if is_server:
                        self.os.version = '2012'
                    else:
                        self.os.version = '8'
                elif self.os.kernel_version.startswith('6.3'):
                    self.os.version = '8.1'
        return

    def _compute_nb_cores(self):
        if self.os.name == "linux":
            try:
                with open('/proc/cpuinfo') as f:
                    return len([1 for line in f if 'processor' in line])
            except IOError:
                return UNKNOWN
        elif self.os.name == 'windows':
            return int(os.environ.get('NUMBER_OF_PROCESSORS', 1))
        elif self.os.name == 'darwin':
            return int(Run(['/usr/sbin/sysctl', '-n', 'hw.ncpu']).out)
        elif self.os.name.lower() == 'freebsd':
            return int(Run(['/sbin/sysctl', '-n', 'hw.ncpu']).out)
        elif self.os.name == 'aix':
            # Try to parse the output of lscfg to get the number of processors
            lscfg_env = os.environ.copy()
            lscfg_env['LANG'] = 'C'
            return len([line for line in
                        Run(['lscfg'], env=lscfg_env).out.splitlines()
                        if line.rstrip().endswith('Processor')])
        elif self.os.name == 'solaris':
            # First get the output of kstat cpu_info
            try:
                cpu_info = Run(['kstat', 'cpu_info']).out
            except OSError:
                # This can happen if kstat is not found.  Do not crash,
                # just return UNKNOWN.
                return UNKNOWN

            # Then try to get the number of core_id
            core_ids = set((line for line in cpu_info.splitlines()
                            if 'core_id' in line))
            nb_cores = len(core_ids)

            # If empty get the number of chip_id
            if nb_cores == 0:
                chip_ids = set((line for line in cpu_info.splitlines()
                                if 'chip_id' in line))
                nb_cores = len(chip_ids)

            return nb_cores
        else:
            return UNKNOWN


if __name__ == "__main__":
    print Arch()
