"""Decorators

to enable/disable a function,
to memoize a function results...
"""

from functools import partial


def enabled(func):
    """This decorator doesn't add any behavior

    >>> @enabled
    ... def foo():
    ...     print "I'm foo"
    >>> foo()
    I'm foo
    """
    return func


def disabled(func):
    """This decorator disables the provided function, and does nothing

    >>> @disabled
    ... def foo():
    ...     print "I'm foo"
    >>> foo()
    """

    def empty_func(*args, **kargs):
        pass
    return empty_func


class memoize(object):
    """@memoize decorator

    Avoid repeating the calculation of results for previously-processed
    inputs.

    >>> import random
    >>> @memoize
    ... def long_computation(r):
    ...     return random.random()
    >>> k = long_computation(42)
    >>> l = long_computation(42)
    >>> k == l
    True
    >>> j = long_computation(666)
    >>> k == j
    False
    """

    def __init__(self, func):
        self.func = func
        self.cache = {}

    def __call__(self, *args):
        """Return the cache value if exist, else call func"""
        try:
            return self.cache[args]
        except KeyError:
            value = self.func(*args)
            self.cache[args] = value
            return value
        except TypeError:
            # non-hashable arguments, skip the cache
            return self.func(*args)

    def __repr__(self):
        """Return the function's docstring."""
        return self.func.__doc__

    def __get__(self, obj, objtype):
        """Support instance methods."""
        return partial(self.__call__, obj)
