 ############################################################################
 #                                                                          #
 #                            REMOTE.PY                                     #
 #                                                                          #
 #           Copyright (C) 2008 - 2014 Ada Core, Inc.                       #
 #                                                                          #
 # This program is free software: you can redistribute it and/or modify     #
 # it under the terms of the GNU General Public License as published by     #
 # the Free Software Foundation, either version 3 of the License, or        #
 # (at your option) any later version.                                      #
 #                                                                          #
 # This program is distributed in the hope that it will be useful,          #
 # but WITHOUT ANY WARRANTY; without even the implied warranty of           #
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
 # GNU General Public License for more details.                             #
 #                                                                          #
 # You should have received a copy of the GNU General Public License        #
 # along with this program.  If not, see <http://www.gnu.org/licenses/>     #
 #                                                                          #
 ############################################################################

"""This module provides simple functions to perform ssh and scp

If module paramiko is loaded then it offers paramiko-based functions which
allows better compatibility with unix and windows systems.
Otherwise it uses gnatpython.ex.Run implementation.
"""
from gnatpython.fileutils import unixpath
from gnatpython.ex import Run
import logging
import os


class SSHError(Exception):
    pass

try:
    import paramiko
    paramiko_logger = logging.getLogger('paramiko')
    paramiko_logger.setLevel(logging.ERROR)

    class SSHClient(paramiko.SSHClient):
        """Override exec_command to add a timeout"""
        def exec_command(self, command, bufsize=-1, timeout=None):
            chan = self._transport.open_session()
            chan.settimeout(timeout)
            chan.exec_command(command)
            stdout = chan.makefile('rb', bufsize)
            stderr = chan.makefile_stderr('rb', bufsize)
            status = chan.recv_exit_status()
            return (stdout, stderr, status)

    def ssh_exec(user, host, cmd, timeout=None):
        """Execute a command on the SSH server

        RETURN:
          The command's (output, error, ecit status) streams as a string tuple.

        PARAMETERS:
          user: user login to authenticate as
          host: the server to connect to
          cmd: the command to execute
          timeout: seconds to wait before raising socket.timeout
        """
        client = SSHClient()
        client.load_system_host_keys()
        client.connect(host, username=user)
        stdout, stderr, status = client.exec_command(cmd, timeout=timeout)
        response = (stdout.read().rstrip(), stderr.read().rstrip(), status)
        client.close()
        return response

    def scp(user, host, remotepath, localpath, method=None):
        client = SSHClient()
        client.load_system_host_keys()
        client.connect(host, username=user)
        sftp = client.open_sftp()
        out = ''
        err = ''
        status = ''
        remotepath = unixpath(remotepath)
        localpath = unixpath(localpath)
        try:
            if method == 'get':
                out = localpath
                status = 0
                sftp.get(remotepath, localpath)
            elif method == 'put':
                out = remotepath
                status = 0
                sftp.put(localpath, remotepath)
            else:
                raise SSHError("Method %s not implemented:" % method)
        except IOError, e:
            err = e
            status = 1
        except SSHError, e:
            err = e
            status = 1
        finally:
            sftp.close()
            client.close()
        return (out, err, status)

    def scp_get(user, host, remotepath, localpath=None):
        """Copy a remote file from host to the local host as localpath

        RETURN:
          (out, err, status) where out is the localpath

        PARAMETERS:
          user: user login to authenticate as
          host: the server to connect to
          remotepath: the remote file to copy (str)
          localpath: the destination path on the local host (str)
        """
        if localpath is None:
            localpath = os.path.basename(remotepath)
        return scp(user=user, host=host, remotepath=remotepath,
                   localpath=localpath, method='get')

    def scp_put(user, host,
                localpath, remotepath=None):
        """Copy a local file to the host as remotepath

        RETURN:
          (out, err, status): where out is the remotepath

        PARAMETERS:
          user: user login to authenticate as
          host: the server to connect to
          localpath: the local file to copy to the host
          remotepath: the remote file.
        """
        if remotepath is None:
            remotepath = os.path.basename(localpath)
        return scp(user=user, host=host, remotepath=remotepath,
                   localpath=localpath, method='put')

except ImportError:

    def ssh_exec(user, host, cmd, timeout=None):
        """Execute a command on the SSH server

        RETURN:
          The command's (input, output, error) streams as a string tuple.

        PARAMETERS:
          user: user login to authenticate as
          host: the server to connect to
          cmd: the command to execute
        """
        ssh = Run(['ssh', '-n', '-q',
                   '-o', 'StrictHostKeyChecking=no',
                   '-o', 'PasswordAuthentication=no',
                   '-l', user, host, cmd], set_sigpipe=False,
                  timeout=timeout)
        return (ssh.out, ssh.err, ssh.status)

    def scp_get(user, host, remotepath, localpath=None):
        """Copy a remote file from host to the local host as localpath

        RETURN:
          (out, err, status) where out is the localpath

        PARAMETERS:
          user: user login to authenticate as
          host: the server to connect to
          remotepath: the remote file to copy (str)
          localpath: the destination path on the local host (str)
        """
        if localpath is None:
            localpath = os.path.basename(remotepath)
        localpath = unixpath(localpath)
        remotepath = unixpath(remotepath)
        p = Run(['scp', '-B', '%s@%s:%s' % (user, host, remotepath),
                 localpath])
        return (localpath, p.err, p.status)

    def scp_put(user, host, localpath, remotepath=None):
        """Copy a local file to the host as remotepath

        RETURN:
          (out, err, status): where out is the remotepath

        PARAMETERS:
          user: user login to authenticate as
          host: the server to connect to
          localpath: the local file to copy to the host
          remotepath: the remote file.
        """
        if remotepath is None:
            remotepath = os.path.basename(localpath)
        p = Run(['scp', '-B', unixpath(localpath),
                 '%s@%s:%s' % (user, host, unixpath(remotepath))])
        return (remotepath, p.err, p.status)
